# Activity functions require no special implementation aside from standard Azure Functions
#   unit testing for Python. As such, no test is implemented here.
# For more information about testing Azure Functions in Python, see the official documentation:
# https://learn.microsoft.com/en-us/azure/azure-functions/functions-reference-python#unit-testing


import datetime
import re
from typing import Dict, Any

from ...constants import DATETIME_STRING_FORMAT


def add_attrib(json_dict: Dict[str, Any], object_,
               attribute_name: str, alt_name: str = None):
    """Add the value of the attribute from the object to the dictionary.

    Used to dynamically add the value of the attribute if the value is present.

    Parameters
    ----------
    json_dict: The dictionary to add the attribute to
    object_: The object to look for the attribute on
    attribute_name: The name of the attribute to look for
    alt_name: An alternate name to provide to the attribute in the in the dictionary
    """
    if hasattr(object_, attribute_name):
        json_dict[alt_name or attribute_name] = \
            getattr(object_, attribute_name)


def add_datetime_attrib(json_dict: Dict[str, Any], object_,
                        attribute_name: str, alt_name: str = None):
    """Add the value of the attribute from the object to the dictionary converted into a string.

    Parameters
    ----------
    json_dict: The dictionary to add the attribute to
    object_: The object to look for the attribute on
    attribute_name: The name of the attribute to look for
    alt_name: An alternate name to provide to the attribute in the in the dictionary
    """
    if hasattr(object_, attribute_name):
        json_dict[alt_name or attribute_name] = \
            getattr(object_, attribute_name).strftime(DATETIME_STRING_FORMAT)


# When we recieve properties from WebJobs extension originally parsed as
#   TimeSpan objects through Newtonsoft, the format complies with the constant
#   format specifier for TimeSpan in .NET.
#   Python offers no convenient way to parse these back into timedeltas,
#   so we use this regex method instead
def parse_timespan_attrib(from_str: str) -> datetime.timedelta:
    """Convert a string representing TimeSpan.ToString("c") in .NET to a python timedelta.

    Parameters
    ----------
    from_str: The string format of the TimeSpan to convert

    Returns
    -------
    timespan.timedelta
        The TimeSpan expressed as a Python datetime.timedelta

    """
    match = re.match(r"^(?P<negative>-)?(?:(?P<days>[0-9]*)\.)?"
                     r"(?P<hours>[0-9]{2}):(?P<minutes>[0-9]{2})"
                     r":(?P<seconds>[0-9]{2})(?:\.(?P<ticks>[0-9]{7}))?$",
                     from_str)
    if match:
        groups = match.groupdict()
        span = datetime.timedelta(
            days=int(groups['days'] or "0"),
            hours=int(groups['hours']),
            minutes=int(groups['minutes']),
            seconds=int(groups['seconds']),
            microseconds=int(groups['ticks'] or "0") // 10)

        if groups['negative'] == '-':
            span = -span
        return span
    else:
        raise Exception(f"Format of TimeSpan failed attempted conversion to timedelta: {from_str}")


def add_json_attrib(json_dict: Dict[str, Any], object_,
                    attribute_name: str, alt_name: str = None):
    """Add the results of the to_json() function call of the attribute from the object to the dict.

    Used to dynamically add the JSON converted value of the attribute if the value is present.

    Parameters
    ----------
    json_dict: The dictionary to add the attribute to
    object_: The object to look for the attribute on
    attribute_name: The name of the attribute to look for
    alt_name: An alternate name to provide to the attribute in the in the dictionary
    """
    if hasattr(object_, attribute_name):
        attribute_value = getattr(object_, attribute_name)
        if attribute_value:
            json_dict[alt_name or attribute_name] = attribute_value.to_json()



from azure.durable_functions.models.ReplaySchema import ReplaySchema
import json
import pytest
from typing import Dict

from azure.durable_functions.constants import HTTP_ACTION_NAME
from azure.durable_functions.models import DurableHttpRequest
from .orchestrator_test_utils import assert_orchestration_state_equals, \
    get_orchestration_state_result, assert_valid_schema, assert_dict_are_equal
from tests.test_utils.ContextBuilder import ContextBuilder
from azure.durable_functions.models.OrchestratorState import OrchestratorState
from azure.durable_functions.models.actions.CallHttpAction import CallHttpAction
from azure.durable_functions.models.TokenSource import ManagedIdentityTokenSource

TEST_URI: str = \
    'https://localhost:7071/we_just_need_a_uri_to_use_for_testing'
SIMPLE_RESULT: str = json.dumps({'name': 'simple'})
CONTENT = json.dumps({'name': 'some data', 'additional': 'data'})
HEADERS = {'header1': 'value1', 'header2': 'value2'}
TOKEN_SOURCE = ManagedIdentityTokenSource('https://management.core.windows.net/')


def simple_get_generator_function(context):
    url = TEST_URI
    yield context.call_http("GET", url)


def complete_generator_function(context):
    url = TEST_URI

    yield context.call_http(method="POST", uri=url, content=json.loads(CONTENT),
                            headers=HEADERS, token_source=TOKEN_SOURCE)


def base_expected_state(output=None, replay_schema: ReplaySchema = ReplaySchema.V1) -> OrchestratorState:
    return OrchestratorState(is_done=False, actions=[], output=output, replay_schema=replay_schema.value)


def add_http_action(state: OrchestratorState, request):
    action = CallHttpAction(request)
    state.actions.append([action])


def add_completed_http_events(
        context_builder: ContextBuilder, id_: int, result: str):
    context_builder.add_task_scheduled_event(name=HTTP_ACTION_NAME, id_=id_)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_task_completed_event(id_=id_, result=result)


def add_failed_http_events(
        context_builder: ContextBuilder, id_: int, reason: str, details: str):
    context_builder.add_task_scheduled_event(name=HTTP_ACTION_NAME, id_=id_)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_task_failed_event(
        id_=id_, reason=reason, details=details)


def get_request() -> DurableHttpRequest:
    return DurableHttpRequest(method='GET', uri=TEST_URI)


def post_request() -> DurableHttpRequest:
    return DurableHttpRequest(method="POST", uri=TEST_URI, content=json.dumps(CONTENT),
                              headers=HEADERS, token_source=TOKEN_SOURCE)


def test_initial_orchestration_state():
    context_builder = ContextBuilder('test_simple_function')

    result = get_orchestration_state_result(
        context_builder, simple_get_generator_function)

    expected_state = base_expected_state()
    request = get_request()
    add_http_action(expected_state, request)
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_completed_state():
    context_builder = ContextBuilder('test_simple_function')
    add_completed_http_events(context_builder, 0, SIMPLE_RESULT)

    result = get_orchestration_state_result(
        context_builder, simple_get_generator_function)

    expected_state = base_expected_state()
    request = get_request()
    add_http_action(expected_state, request)
    expected_state._is_done = True
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_failed_state():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_failed_http_events(
        context_builder, 0, failed_reason, failed_details)

    try:
        result = get_orchestration_state_result(
            context_builder, simple_get_generator_function)
        # We expected an exception
        assert False
    except Exception as e:
        error_label = "\n\n$OutOfProcData$:"
        error_str = str(e)

        expected_state = base_expected_state()
        request = get_request()
        add_http_action(expected_state, request)

        error_msg = f'{failed_reason} \n {failed_details}'
        expected_state._error = error_msg
        state_str = expected_state.to_json_string()
        
        expected_error_str = f"{error_msg}{error_label}{state_str}"
        assert expected_error_str == error_str


def test_initial_post_state():
    context_builder = ContextBuilder('test_simple_function')

    result = get_orchestration_state_result(
        context_builder, complete_generator_function)

    expected_state = base_expected_state()
    request = post_request()
    add_http_action(expected_state, request)
    expected = expected_state.to_json()

    # assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)
    validate_result_http_request(result)


def validate_result_http_request(result):
    http_request = result['actions'][0][0]['httpRequest']
    assert http_request is not None
    assert http_request['method'] == 'POST'
    assert http_request['uri'] == TEST_URI
    content = http_request['content']
    assert isinstance(content, str)
    content = json.loads(content)
    test_content = json.loads(CONTENT)
    assert_dict_are_equal(test_content, content)
    assert content['name'] == 'some data'
    headers: Dict[str, str] = http_request['headers']
    assert_dict_are_equal(HEADERS, headers)
    assert http_request['tokenSource']['resource'] == TOKEN_SOURCE.resource


def test_post_completed_state():
    context_builder = ContextBuilder('test_simple_function')
    add_completed_http_events(context_builder, 0, SIMPLE_RESULT)

    result = get_orchestration_state_result(
        context_builder, complete_generator_function)

    expected_state = base_expected_state()
    request = post_request()
    add_http_action(expected_state, request)
    expected_state._is_done = True
    expected = expected_state.to_json()

    # assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)
    validate_result_http_request(result)

@pytest.mark.parametrize("content, expected_content, is_raw_str", [
    (None, None, False),
    ("string data", '"string data"', False),
    ('{"key": "value"}', '"{\\"key\\": \\"value\\"}"', False),
    ('["list", "content"]', '"[\\"list\\", \\"content\\"]"', False),
    ('[]', '"[]"', False),
    ('42', '"42"', False),
    ('true', '"true"', False),  
    # Cases that test actual behavior (not strictly adhering to Optional[str])
    ({"key": "value"}, '{"key": "value"}', False),
    (["list", "content"], '["list", "content"]', False),
    ([], '[]', False),
    (42, '42', False),
    (True, 'true', False),
    # Cases when is_raw_str is True
    ("string data", "string data", True),
    ('{"key": "value"}', '{"key": "value"}', True),
    ('[]', '[]', True),
])
def test_call_http_content_handling(content, expected_content, is_raw_str):
    def orchestrator_function(context):
        yield context.call_http("POST", TEST_URI, content, is_raw_str=is_raw_str)

    context_builder = ContextBuilder('test_call_http_content_handling')
    result = get_orchestration_state_result(context_builder, orchestrator_function)

    assert len(result['actions']) == 1
    http_action = result['actions'][0][0]['httpRequest']
    
    assert http_action['method'] == "POST"
    assert http_action['uri'] == TEST_URI
    assert http_action['content'] == expected_content

# Test that call_http raises a TypeError when is_raw_str is True but content is not a string
def test_call_http_non_string_content_with_raw_str():
    def orchestrator_function(context):
        yield context.call_http("POST", TEST_URI, {"key": "value"}, is_raw_str=True)

    context_builder = ContextBuilder('test_call_http_non_string_content_with_raw_str')
    
    try:
        result = get_orchestration_state_result(context_builder, orchestrator_function)
        assert False
    except Exception as e:
        error_label = "\n\n$OutOfProcData$:"
        error_str = str(e)

        expected_state = base_expected_state()
        error_msg = "Invalid use of 'is_raw_str' parameter: 'is_raw_str' is "\
                "set to 'True' but 'content' is not an instance of type 'str'. "\
                "Either set 'is_raw_str' to 'False', or ensure your 'content' "\
                "is of type 'str'."
        expected_state._error = error_msg
        state_str = expected_state.to_json_string()
        
        expected_error_str = f"{error_msg}{error_label}{state_str}"
        assert expected_error_str == error_str


import pytest
from azure.durable_functions.models.utils.entity_utils import EntityId

@pytest.mark.parametrize(
    ("name_e1", "key_1", "name_e2", "key_2", "expected"),
    [
        ("name1", "key1", "name1", "key1", True),
        ("name1", "key1", "name1", "key2", False),
        ("name1", "key1", "name2", "key1", False),
        ("name1", "key1", "name2", "key2", False),
    ],
)
def test_equal_entity_by_name_and_key(name_e1, key_1, name_e2, key_2, expected):

    entity1 = EntityId(name_e1, key_1)
    entity2 = EntityId(name_e2, key_2)

    assert (entity1 == entity2) == expected

def test_equality_with_non_entity_id():

    entity = EntityId("name", "key")

    assert (entity == "not an entity id") == False



import unittest
from unittest.mock import Mock, call, patch
from azure.durable_functions.testing import orchestrator_generator_wrapper

from function_app import my_orchestrator


@patch('azure.durable_functions.models.TaskBase')
def mock_activity(activity_name, input, task):
  if activity_name == "say_hello":
    task.result = f"Hello {input}!"
    return task
  raise Exception("Activity not found")


class TestFunction(unittest.TestCase):
  @patch('azure.durable_functions.DurableOrchestrationContext')
  def test_chaining_orchestrator(self, context):
    # Get the original method definition as seen in the function_app.py file
    func_call = my_orchestrator.build().get_user_function().orchestrator_function

    context.call_activity = Mock(side_effect=mock_activity)

    # Create a generator using the method and mocked context
    user_orchestrator = func_call(context)

    # Use orchestrator_generator_wrapper to get the values from the generator.
    # Processes the orchestrator in a way that is equivalent to the Durable replay logic
    values = [val for val in orchestrator_generator_wrapper(user_orchestrator)]

    expected_activity_calls = [call('say_hello', 'Tokyo'),
                               call('say_hello', 'Seattle'),
                               call('say_hello', 'London')]

    self.assertEqual(context.call_activity.call_count, 3)
    self.assertEqual(context.call_activity.call_args_list, expected_activity_calls)
    self.assertEqual(values[3], ["Hello Tokyo!", "Hello Seattle!", "Hello London!"])
